# -*- coding: utf-8 -*-

from .chord import as_chord, Chord


class ChordProgression(object):
    """ Class to handle chord progressions.

    :type initial_chords: str|pychord.Chord|list
    :param initial_chords: Initial chord or chords of the chord progressions
    """

    def __init__(self, initial_chords=None):
        if initial_chords is None:
            initial_chords = []
        if isinstance(initial_chords, Chord):
            self.chords = [initial_chords]
        elif isinstance(initial_chords, str):
            self.chords = [as_chord(initial_chords)]
        elif isinstance(initial_chords, list):
            self.chords = [as_chord(chord) for chord in initial_chords]
        else:
            raise TypeError("Cannot initialize ChordProgression with argument of {} type".format(type(initial_chords)))

    def __unicode__(self):
        return " | ".join([chord.chord for chord in self.chords])

    def __str__(self):
        return " | ".join([chord.chord for chord in self.chords])

    def __add__(self, other):
        self.chords += other.chords
        return self

    def __len__(self):
        return len(self.chords)

    def __getitem__(self, item):
        return self.chords[item]

    def get_chords(self):
        return self.chords

    def append(self, chord):
        """ Append a chord to chord progressions

        :type chord: str|pychord.Chord
        :param chord: A chord to append
        :return:
        """
        self.chords.append(as_chord(chord))

    def insert(self, index, chord):
        """ Insert a chord to chord progressions

        :param int index: Index to insert a chord
        :type chord: str|pychord.Chord
        :param chord: A chord to insert
        :return:
        """
        self.chords.insert(index, as_chord(chord))

    def pop(self, index=-1):
        """ Pop a chord from chord progressions

        :param int index: Index of the chord to pop (default: -1)
        :return pychord.Chord:
        """
        return self.chords.pop(index)

    def transpose(self, trans):
        """ Transpose whole chord progressions

        :param int trans: Transpose key
        :return:
        """
        for chord in self.chords:
            chord.transpose(trans)



# -*- coding: utf-8 -*-


from .parser import parse
from .utils import transpose_note, display_appended, display_on


class Chord(object):
    """ Class to handle a chord.

    :param str chord: Name of the chord. (e.g. C, Am7, F#m7-5/A)
    """
    def __init__(self, chord):
        self.chord = chord
        self.root, self.quality, self.appended, self.on = "", "", "", ""
        self._parse(chord)

    def __unicode__(self):
        return "Chord({})".format(self.chord)

    def __str__(self):
        return "Chord({})".format(self.chord)

    def __eq__(self, other):
        return self.chord == other.chord

    def info(self):
        return """{}
root={}
quality={}
appended={}
on={}""".format(self.chord, self.root, self.quality, self.appended, self.on)

    def transpose(self, trans):
        """ Transpose the chord

        :param int trans: Transpose key
        :return:
        """
        if not isinstance(trans, int):
            raise TypeError("Expected integers, not {}".format(type(trans)))
        self.root = transpose_note(self.root, trans)
        if self.on:
            self.on = transpose_note(self.on, trans)
        self._reconfigure_chord()

    def components(self, visible=True):
        return self.quality.get_components(root=self.root, visible=visible)

    def _parse(self, chord):
        root, quality, appended, on = parse(chord)
        self.root = root
        self.quality = quality
        self.appended = appended
        self.on = on

    def _reconfigure_chord(self):
        # TODO: Use appended
        self.chord = "{}{}{}{}".format(self.root,
                                       self.quality.quality,
                                       display_appended(self.appended),
                                       display_on(self.on))


def as_chord(chord):
    """ convert from str to Chord instance if input is str

    :type chord: str|pychord.Chord
    :param chord: Chord name or Chord instance
    :return pychord.Chord: Chord instance
    """
    if isinstance(chord, Chord):
        return chord
    elif isinstance(chord, str):
        return Chord(chord)
    else:
        raise TypeError("input type should be str or Chord instance.")


# -*- coding: utf-8 -*-

from .constants import QUALITY_DICT
from .utils import note_to_val, val_to_note


class Quality(object):
    """ Chord quality

    :param str quality: str expression of chord quality
    """
    def __init__(self, quality):
        if quality not in QUALITY_DICT:
            raise ValueError("unknown quality {}".format(quality))
        self.quality = quality
        self.components = QUALITY_DICT[quality]

    def __unicode__(self):
        return self.quality

    def __str__(self):
        return self.quality

    def get_components(self, root='C', visible=False):
        """ get components of chord quality

        :param str root: the root note of the chord
        :param bool visible: returns the name of notes if True
        :return list:
        """
        root_val = note_to_val(root)
        components = [v + root_val for v in self.components]
        if visible:
            components = [val_to_note(c, scale=root) for c in components]
        return components

    def append_on_chord(self, on_chord, root):
        """ Append on chord

        To create Am7/G
        q = Quality('m7')
        q.append_on_chord('G', root='A')

        :param str on_chord: bass note of the chord
        :param str root: root note of the chord
        """
        root_val = note_to_val(root)
        on_chord_val = note_to_val(on_chord) - root_val
        for idx, val in enumerate(self.components):
            if val % 12 == on_chord_val:
                scale = val / 12
                self._rotate_components(idx, scale)
                break
        if on_chord_val > root_val:
            on_chord_val -= 12
        if on_chord_val not in self.components:
            self.components.insert(0, on_chord_val)

    def _rotate_components(self, stop_idx, scale):
        for idx, val in enumerate(self.components[:stop_idx]):
            self.components[idx] += (scale + 1) * 12

    def append_note(self, note, root, scale=0):
        root_val = note_to_val(root)
        note_val = note_to_val(note) - root_val + scale * 12
        if note_val not in self.components:
            self.components.append(note_val)
            self.components.sort()

    def append_notes(self, notes, root, scale=0):
        for note in notes:
            self.append_note(note, root, scale)
